(load (merge-pathnames "../common.lisp" *load-truename*))

(defconstant test-data
  '(
    "ecl:gry pid:860033327 eyr:2020 hcl:#fffffd"
    "byr:1937 iyr:2017 cid:147 hgt:183cm"
    ""
    "iyr:2013 ecl:amb cid:350 eyr:2023 pid:028048884"
    "hcl:#cfa07d byr:1929"
    ""
    "hcl:#ae17e1 iyr:2013"
    "eyr:2024"
    "ecl:brn pid:760753108 byr:1931"
    "hgt:179cm"
    ""
    "hcl:#cfa07d eyr:2025 pid:166559648"
    "iyr:2011 ecl:brn hgt:59in"))

(defun all-chars (str p)
  (loop for i across str
        when (not (funcall p i)) return NIL
          finally (return T)))

(defun validation-present (x) (not (null x)))

(defun validation-in-range (x from to)
  (let ((value (parse-integer x :junk-allowed T)))
    (and value
         (<= from value)
         (<= value to))))

(defun validation-byr (x)
  (validation-in-range x 1920 2002))

(defun validation-iyr (x)
  (validation-in-range x 2010 2020))

(defun validation-eyr (x)
  (validation-in-range x 2020 2030))

(defun validation-hcl (x)
  (and (char= #\# (aref x 0))
       (= 7 (length x))
       (all-chars (subseq x 1 nil) (lambda (x) (digit-char-p x 16)))))

(defun validation-pid (x)
  (and (= 9 (length x))
       (all-chars x #'digit-char-p)))

(defun validation-hgt (x)
  (multiple-value-bind (n j) (parse-integer x :junk-allowed T)
    (let ((unit (subseq x j nil)))
      (cond
        ((string= "in" unit) (and (<= 59 n) (<= n 76)))
        ((string= "cm" unit) (and (<= 150 n) (<= n 193)))))))

(defun validation-ecl (x)
  (member x '("amb" "blu" "brn" "gry" "grn" "hzl" "oth") :test #'string=))

(defconstant requirements
  (list
   (cons "byr" #'validation-byr) ;; (Birth Year)
   (cons "iyr" #'validation-iyr) ;; (Issue Year)
   (cons "eyr" #'validation-eyr) ;; (Expiration Year)
   (cons "hgt" #'validation-hgt) ;; (Height)
   (cons "hcl" #'validation-hcl) ;; (Hair Color)
   (cons "ecl" #'validation-ecl) ;; (Eye Color)
   (cons "pid" #'validation-pid) ;; (Passport ID)
   ;; "cid" ;; (Country ID)
   ))

(defun parse-kv (part)
  (let ((split (position #\: part)))
    (cons (subseq part 0 split) (subseq part (1+ split) nil))))

(defun parse-input (lines)
  (loop for chunk in lines
        with current = '()
        if (string= chunk "")
          collect (let ((result current))
                    (setf current '())
                    result)
        else
          do (loop for x = 0 then (1+ j)
                   as j = (position #\Space chunk :start x :test #'char=)
                   do (setf current (cons (parse-kv (subseq chunk x j)) current))
                   while j)
        end))

(defun password-valid (validations password)
  (if (null validations)
      T
      (if (validate (car validations) password)
          (password-valid (cdr validations) password))))

(defun part1 (input)
  (let ((validations (mapcar (lambda (x) (cons (car x) #'validation-present)) requirements)))
    (loop for passport in (parse-input input)
          counting (password-valid validations passport))))

(defun part2 (input)
  (loop for passport in (parse-input input)
        counting (password-valid requirements passport)))

(defun validate (what alist)
  (let ((value (cdr (assoc (car what) alist :test #'string=))))
    (when value (funcall (cdr what) value))))

(assert (= 2 (part1 test-data)))

(assert (validation-byr "2002"))
(assert (not (validation-byr "2003")))
(assert (validation-ecl "brn"))
(assert (not (validation-ecl "wat")))
(assert (validation-hcl "#123abc"))
(assert (not (validation-hcl "#123abz")))
(assert (not (validation-hcl "123abc")))
(assert (validation-pid "000000001"))
(assert (not (validation-pid "0123456789")))
(assert (validation-hgt "60in"))
(assert (validation-hgt "190cm"))
(assert (not (validation-hgt "190in")))
(assert (not (validation-hgt "190")))

(let ((input (read-data)))
  (format t "Part 1: ~A~%" (part1 input))
  (format t "Part 2: ~A~%" (part2 input)))
