#lang racket

(provide part1 part2)

(require threading)

(define (part1 input)
  (define (address-map mask addr) (list addr))
  (solution input address-map mask-number))

(define (part2 input)
  (define (address-map mask addr)
    (~>> addr
         number->bitstring
         (apply-mask-address mask)
         (apply cartesian-product)
         (map (compose bitstring->number list->string))))
  (define (value-map mask value) value)
  (solution input address-map value-map))

(define (solution input address-map value-map)
  (for/sum ([v (in-hash-values (evaluate input address-map value-map))]) v))

(define (evaluate program address-map value-map)
  (for/fold ([mask (number->bitstring 0)]
             [memory (hash)]
             #:result memory)
      ([line (in-list program)])

    (cond
     [(regexp-match? mask-re line)
      (values (parse-mask line) memory)]
     [else
      (define-values (addr value) (parse-assign line))
      (values mask (write-into-addresses memory
                                         (address-map mask addr)
                                         (value-map mask value)))])))

(define mask-re #px"mask = ([X01]{36})")

(define assign-re #px"mem\\[([0-9]+)\\] = ([0-9]+)")

(define (number->bitstring number)
  (~a (number->string number 2) #:width 36 #:align 'right #:pad-string "0"))

(define (bitstring->number number)
  (string->number number 2))

(define (mask-number mask number)
  (~>> number
       number->bitstring
       (apply-mask mask)
       bitstring->number))

(define (apply-mask mask bitstring)
  (list->string
   (for/list ([m mask]
              [s bitstring])
     (if (char=? m #\X) s m))))

(define (apply-mask-address mask bitstring)
  (for/list ([m mask]
             [s bitstring])
    (case m
      [(#\0) (list s)]
      [(#\1) (list #\1)]
      [(#\X) (list #\0 #\1)])))

(define (parse-mask line)
  (cadr (regexp-match mask-re line)))

(define (parse-assign line)
  (~>> (regexp-match assign-re line)
       cdr
       (map string->number)
       (apply values)))

(define (write-into-addresses memory addrs value)
  (for/fold ([memory memory])
      ([addr (in-list addrs)])
    (hash-set memory addr value)))

(define (compose f1 f2)
  (lambda (x) (f1 (f2 x))))

(module+ test
  (require rackunit)
  (define test-data
    '("mask = XXXXXXXXXXXXXXXXXXXXXXXXXXXXX1XXXX0X"
      "mem[8] = 11"
      "mem[7] = 101"
      "mem[8] = 0"))

  (check-equal? (parse-mask "mask = XXXXXXXXXXXXXXXXXXXXXXXXXXXXX1XXXX0X")
                "XXXXXXXXXXXXXXXXXXXXXXXXXXXXX1XXXX0X")

  (let-values ([(addr value) (parse-assign "mem[7] = 101")])
    (check-equal? addr 7)
    (check-equal? value 101))

  (let ([mask "XXXXXXXXXXXXXXXXXXXXXXXXXXXXX1XXXX0X"])
    (check-equal? (apply-mask mask "000000000000000000000000000000001011")
                  "000000000000000000000000000001001001")
    (check-equal? (apply-mask mask "000000000000000000000000000001100101")
                  "000000000000000000000000000001100101")
    (check-equal? (apply-mask mask "000000000000000000000000000000000000")
                  "000000000000000000000000000001000000"))

  (check-equal? 165 (part1 test-data))

  (define test-program-2
    '("mask = 000000000000000000000000000000X1001X"
      "mem[42] = 100"
      "mask = 00000000000000000000000000000000X0XX"
      "mem[26] = 1"))

  (check-equal? 208 (part2 test-program-2)))
