open System.Text.RegularExpressions

let (|Regex|_|) pattern input =
    let m = Regex.Match(input, pattern)

    if m.Success then
        Some(List.tail [ for g in m.Groups -> g.Value ])
    else
        None

type Operation =
    | Square
    | Multiply of bigint
    | Add of bigint

type Monkey =
    { id: int
      mutable items: bigint list
      operation: Operation
      test: bigint
      ifTrue: int
      ifFalse: int
      mutable inspectedItems: bigint }

let parseBlock lines =
    let [| id; items; op; test; ift; iff |] = lines

    let parseNumber parse x =
        match x with
        | Regex @"(\d+)" [ n ] -> parse n
        | _ -> failwith "unreachable"

    let items =
        match items with
        | Regex @" ((\d+)(, (\d+))*)" (x :: rest) -> x.Split(", ") |> Seq.map (int >> bigint)
        | _ -> failwith "unreachable"

    let operation =
        match op with
        | Regex @"new = old \* (\d+)" [ n ] -> Multiply(bigint (int n))
        | Regex @"new = old \+ (\d+)" [ n ] -> Add(bigint (int n))
        | Regex @"new = old \* old" [] -> Square
        | _ -> failwith "unreachable"

    { id = parseNumber int id
      items = items |> List.ofSeq
      operation = operation
      test = parseNumber (int >> bigint) test
      ifTrue = parseNumber int ift
      ifFalse = parseNumber int iff
      inspectedItems = 0I }

let parse lines =
    lines
    |> Seq.filter (fun l -> not (l = ""))
    |> Seq.chunkBySize 6
    |> Seq.map parseBlock

let apply op i =
    match op with
    | Add n -> i + n
    | Multiply n -> i * n
    | Square -> i * i

let solution rounds relieve input =
    let mutable g = Array.ofSeq input

    let rec turn m =
        let monkey = g.[m]

        match monkey.items with
        | [] -> ()
        | i :: rest ->
            let i = i |> apply monkey.operation |> relieve

            let dst =
                if i % monkey.test = 0I then
                    monkey.ifTrue
                else
                    monkey.ifFalse

            g.[m].inspectedItems <- g.[m].inspectedItems + 1I
            g.[m].items <- rest
            g.[dst].items <- g.[dst].items @ [ i ]

            turn m

    for _ in 1..rounds do
        for i in 0 .. Array.length g - 1 do
            turn i

    g
    |> Seq.map (fun m -> m.inspectedItems)
    |> Seq.sort
    |> Seq.rev
    |> Seq.take 2
    |> Seq.reduce (*)

let part1 input = solution 20 (fun x -> x / 3I) input

open System.IO

let parsed = fsi.CommandLineArgs.[1] |> File.ReadAllLines |> parse

#time
let p1 = part1 parsed
#time
printfn "Part 1: %A" p1

#time
let modulo = parsed |> Seq.map (fun x -> x.test) |> Seq.reduce (*)
let part2 = solution 10000 (fun x -> x % modulo)
let p2 = part2 parsed
#time


printfn "Part 2: %A" p2
