open System.Text.RegularExpressions

let (|Regex|_|) pattern input =
    let m = Regex.Match(input, pattern)

    if m.Success then
        Some(List.tail [ for g in m.Groups -> g.Value ])
    else
        None

type Operation =
    | Square
    | Multiply of bigint
    | Add of bigint

type Monkey =
    { id: int
      items: bigint list
      operation: Operation
      test: bigint
      ifTrue: int
      ifFalse: int
      inspectedItems: bigint }

let parseBlock lines =
    let [| id; items; op; test; ift; iff |] = lines

    let parseNumber parse x =
        match x with
        | Regex @"(\d+)" [ n ] -> parse n
        | _ -> failwith "unreachable"

    let items =
        match items with
        | Regex @" ((\d+)(, (\d+))*)" (x :: rest) -> x.Split(", ") |> Seq.map (int >> bigint)
        | _ -> failwith "unreachable"

    let operation =
        match op with
        | Regex @"new = old \* (\d+)" [ n ] -> Multiply(bigint (int n))
        | Regex @"new = old \+ (\d+)" [ n ] -> Add(bigint (int n))
        | Regex @"new = old \* old" [] -> Square
        | _ -> failwith "unreachable"

    { id = parseNumber int id
      items = items |> List.ofSeq
      operation = operation
      test = parseNumber (int >> bigint) test
      ifTrue = parseNumber int ift
      ifFalse = parseNumber int iff
      inspectedItems = bigint 0 }

let parse lines =
    lines
    |> Seq.filter (fun l -> not (l = ""))
    |> Seq.chunkBySize 6
    |> Seq.map parseBlock
    |> Array.ofSeq

let apply op i =
    match op with
    | Add n -> i + n
    | Multiply n -> i * n
    | Square -> i * i

let pushItem m i (g: Monkey[]) =
    let monkey = g.[m]

    g |> Array.updateAt m { monkey with items = monkey.items @ [ i ] }

let solution rounds relieve input =
    let examineItem g m i =
        let i = i |> apply m.operation |> relieve
        let dst = if i % m.test = 0I then m.ifTrue else m.ifFalse

        let _ :: rest = m.items

        let m =
            { m with
                inspectedItems = m.inspectedItems + 1I
                items = rest }

        g |> Array.updateAt m.id m |> pushItem dst i

    let turn g m =
        let rec examineItems' (g: Monkey[]) =
            let monkey = g.[m]

            match monkey.items with
            | [] -> g
            | i :: rest ->
                let g = examineItem g monkey i
                examineItems' g

        examineItems' g

    let runRound g =
        let rec runRound' g i =
            if i >= Array.length g then
                g
            else
                runRound' (turn g i) (i + 1)

        runRound' g 0

    input
    |> Seq.unfold (fun g -> let next = runRound g in Some(next, next))
    |> Seq.take (rounds)
    |> Seq.last
    |> Seq.map (fun m -> m.inspectedItems)
    |> Seq.sort
    |> Seq.rev
    |> Seq.take 2
    |> Seq.reduce (*)

let part1 = solution 20 (fun x -> x / 3I)

open System.IO

let parsed = fsi.CommandLineArgs.[1] |> File.ReadAllLines |> parse

printfn "Part 1: %A" (part1 parsed)

let modulo = parsed |> Seq.map (fun x -> x.test) |> Seq.reduce (*)
let part2 = solution 10000 (fun x -> x % modulo)

printfn "Part 2: %A" (part2 parsed)
